package com.ishland.c2me.opts.scheduling.common;

import com.ishland.c2me.base.common.config.ConfigSystem;

public class Config {

    public static final boolean doMidTickChunkTasks = new ConfigSystem.ConfigAccessor()
            .key("generalOptimizations.doMidTickChunkTasks")
            .comment("""
                    Whether to enable mid-tick chunk tasks \s
                    Mid-tick chunk tasks is to execute chunk tasks during server tick loop \s
                    to speed up chunk loading and generation \s
                    This helps chunks loading and generating under high MSPT but may raise \s
                    MSPT when chunks are loading or generating \s
                    \s
                    Incompatible with Dimensional Threading (dimthread)
                    """)
            .incompatibleMod("dimthread", "*")
            .getBoolean(true, false);

    public static final AutoSaveMode autoSaveMode = new ConfigSystem.ConfigAccessor()
            .key("generalOptimizations.autoSave.mode")
            .comment("""
                    Defines how auto save should be handled \s
                    VANILLA: Use vanilla auto-save behavior (auto-save performed every tick during ticking) \s
                    ENHANCED: Use C2ME enhanced auto-save (auto-save performed when the server have spare time after ticking) \s
                    PERIODIC: Use pre-1.18 vanilla auto-save behavior (auto-save performed every 6000 ticks during ticking) \s
                    \s
                    Please preserve quotes so this config don't break
                    """)
            .getEnum(AutoSaveMode.class, AutoSaveMode.ENHANCED, AutoSaveMode.VANILLA);

    public static final long autoSaveDelayMillis = new ConfigSystem.ConfigAccessor()
            .key("generalOptimizations.autoSave.delay")
            .comment("""
                                Defines the delay in milliseconds between performing auto-save for a chunk \s
                                This is only used when mode is set to ENHANCED
                                """)
            .getLong(20000,20000);

    public static void init() {
    }


    public enum AutoSaveMode {
        VANILLA(false, false),
        ENHANCED(true, true),
        PERIODIC(true, false);

        public final boolean disableVanillaMidTickAutoSave;
        public final boolean enableEnhancedAutoSave;

        AutoSaveMode(boolean disableVanillaMidTickAutoSave, boolean enableEnhancedAutoSave) {
            this.disableVanillaMidTickAutoSave = disableVanillaMidTickAutoSave;
            this.enableEnhancedAutoSave = enableEnhancedAutoSave;
        }
    }

}
